/*									tab:4
 *
 * em3d.h - Header file for the electromagnetic problem in three dimensions.
 *
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Steve Lumetta
 * Version:			11
 * Creation Date:	Sat Jan 23 12:22:55 1993
 * Filename:		em3d.h
 * History:
 *	SL	1	Sat Jan 23 12:42:25 1993
 *		Prepared for release as example (without proper balancing).
 */

#ifndef _EM3D_H
#define _EM3D_H

#undef EXTERN
#undef INIT
#ifdef _EM3D_MAIN
#define EXTERN
#define INIT(a) = (a)
#else
#define EXTERN extern
#define INIT(a)
#endif

#define USE_MCS_LOCKS

#ifdef USE_MCS_LOCKS
#include "mcs_lock.h"

#define lock_decl(l)	MCS_lock_t l
#define lock_init(l)	{ (l) = NULL; }
#define acq_lock(l)	MCS_acquire(&(l), my_qnode)
#define rel_lock(l)	MCS_release(&(l), my_qnode)

EXTERN MCS_q_node_t *my_qnode;

#else

#include "lock.h"

#define lock_decl(l)	LockPtr l
#define lock_init(l)	WWT_LOCKINIT(l)
#define acq_lock(l)	lock(l)
#define rel_lock(l)	unlock(l)

#endif

/* Data type for nodes */
typedef double DATA;

/* List of nodes dependent on a node */
typedef struct to_node_list_t {
    int length;
    struct node_t **nodes;
} to_node_list_t;

/* List of nodes on which a node depends (and coefficients) */
typedef struct from_node_list_t {
    int length;
    DATA* coeffs;
    union {
        struct node_t **nodes;
	DATA **values;
    } nv;
} from_node_list_t;

/* Node structure */
typedef struct node_t {

#ifdef SEPARATE_VALS

    DATA *value_p;
#define node_val(np)	(*((np)->value_p))

#else

    DATA value;
#define node_val(np)	((np)->value)

#endif

    lock_decl(lock);
    struct to_node_list_t to_nodes;
    struct from_node_list_t from_nodes;
} node_t;

/* Prototypes of functions used to obtain original graph */
void all_initialize_graph();

/* Prototypes of functions used to calculate solution */
void all_prepare_packing();
int all_compute_e();
int all_compute_h();

/* Private variables */
EXTERN int n_h_nodes;		/* number of h nodes per processor */
EXTERN int n_e_nodes;		/* number of e nodes per processor */
EXTERN node_t *h_nodes;		/* vector of processor's local h nodes */
EXTERN node_t *e_nodes;		/* vector of processor's local e nodes */

#define MAX_PROCS	32

/* Shared variables */
typedef struct {
    node_t *h_table[MAX_PROCS];
    node_t *e_table[MAX_PROCS];
    
} shmem_t;

EXTERN shmem_t *shmem;

#define MYPROC		wwt_node_number
#define PROCS		wwt_num_nodes
#define barrier()	wwt_cm5_barrier(PROCS)


/* Parameters */
EXTERN int n_nodes INIT(100);
EXTERN int d_nodes INIT(5);
EXTERN int local_p INIT(100);
EXTERN int dist_span INIT(1);
EXTERN int time_steps INIT(100);
EXTERN int verbose INIT(0);

typedef enum {
    RoundRobin,
    Local,
    User
} Version;

EXTERN Version version;	/* no default */

#ifdef USER_PROTOCOL

void *alloc_val_mem(int);

void init_user_protocol();
void start_first_iter();
void end_first_e_iter();
void end_first_h_iter();
void end_e_iter();
void end_h_iter();
void set_h_node_range(void *, void *);
void set_e_node_range(void *, void *);

#else

#define init_user_protocol()
#define start_first_iter()
#define end_first_e_iter()	barrier()
#define end_first_h_iter()	barrier()
#define end_e_iter()		barrier()
#define end_h_iter()		barrier()
#define set_h_node_range(a,b)
#define set_e_node_range(a,b)

#endif

#undef EXTERN
#undef INIT

/*
 * ceil(n/d)
 */
#ifndef DIV_CEIL
#define DIV_CEIL(n,d)   ( ((n)+(d)-1)/(d) )
#endif

#endif

